////////////////////////////////
//
//   Brad Homer
//   Florida State University
//   CAP4601
//   Research Project
//   Summer 2013
//
//   halt.cpp
//
////////////////////////////////

#include <iostream>
#include <fstream>
#include <sstream>
#include <stdexcept>
#include <algorithm>
#include <functional>
#include <cctype>
#include <locale>
#include <string>

using namespace std;

// ltrim
static inline string &ltrim( string &s ) {
   s.erase( s.begin(), find_if( s.begin(), s.end(), not1( ptr_fun<int, int>( isspace ))));
   return s;
}

// rtrim
static inline string &rtrim( string &s ) {
   s.erase( find_if( s.rbegin(), s.rend(), not1( ptr_fun<int, int>(isspace ))).base(), s.end());
   return s;
}

// trim
static inline string &trim( string &s ) {
   return ltrim(rtrim(s));
}

// Will the given code halt or run forever when executed with all of its input?
// Note: headerInput only contains the portion of input which is passed via the header
bool halt(char* codeFileName, char* headerInput[]) {

   ifstream file( codeFileName );
   if( file.fail() )
      throw runtime_error("File cannot be opened.");

   if( !(headerInput == NULL) )
      cout << "Header input unused for these examples.\n";
	  
   string quote = "\"";
   string line;
   bool foundHaltCall = false;
   bool foundNotHaltCall = false;
   bool foundForever = false;
   //bool foundReturn = false;  // not needed for these examples
   bool matchedCall = false;

   while( getline( file, line ) )
   {
      istringstream iss( line );

      string token1;
      string token2;

      // assume headers and syntax allowing for successful compilation are present

      // while handling
      if( trim( line ) == "while(true);" ){
         foundForever = true;
         if (!matchedCall){
            if (foundHaltCall)
               throw runtime_error("Contradiction found: Missing input due to deadlock.");

            if (foundNotHaltCall)
               matchedCall = true;
         }
      }

      // return handling
      if( trim( line ) == "return 0;" ){
         //foundReturn = true;
         if (!matchedCall){
            if (foundHaltCall)
               matchedCall = true;

            if (foundNotHaltCall)
               throw runtime_error("Contradiction found: Missing input due to deadlock.");
         }
      }

      // if handling
      if( getline( iss, token1, '(' )  ) {
         if( trim( token1 ) == "if" ){
            if( getline( iss, token2, '(' ) ){
               if( trim( token2 ) == "halt" ){ // program has conditional logic based on output from halt()
                  if( getline( iss, token1, ',') ){
                     if( trim( token1 ) == (quote + codeFileName + quote) ){
                        foundHaltCall = true;
                     }
                  }
               } else if( trim( token2 ) == "!halt" ){ // program has conditional logic based on negation of output from halt()
                  if( getline( iss, token1, ',') ){
                     if( trim( token1 ) == (quote + codeFileName + quote) ){
                        foundNotHaltCall = true;
                     }
                  }
               }
            }
         }
      }
   } // end while

   if (foundNotHaltCall)
      return false;

   if (foundHaltCall)
      return true;

   if (foundForever)
      return false;
   else
      return true;
}
